;;; -*- Mode: Lisp; Syntax: Common-Lisp; Package: org.langband.vanilla -*-

#|

Based on:

DESC: variants/vanilla/base.lisp - the base variant class for Vanilla
Copyright (c) 2000-2002 - Stig Erik Sand

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

|#

(in-package :org.langband.vanilla)

(defclass icfp-variant (variant)
  ((dawn-time     :initarg :dawntime   :initform 0    :accessor variant.dawn)
   (twilight-time :initarg :twilight   :initform 6000 :accessor variant.twilight)
   (town-seed         :initform nil  :accessor variant.town-seed)
   (used-scroll-names :initform (make-hash-table :test #'equal) :accessor variant.used-scroll-names)
   
   (gold-table     :initarg :gold-table
		   :initform nil
		   :accessor variant.gold-table)
   (spells         :initarg :spells
		   :initform (make-hash-table :test #'equal)
		   :accessor variant.spells)

   (spellbooks     :initarg :spellbooks
		   :initform (make-hash-table :test #'equal)
		   :accessor variant.spellbooks)

   ;; 50 is max-level
   (max-charlevel :initform 50)
   ;; put xp-table here right away
   (xp-table :initform #1A(     10       25       45       70      100
			       140      200      280      380      500
			       650      850     1100     1400     1800
			      2300     2900     3600     4400     5400
			      6800     8400    10200    12500    17500
			     25000    35000    50000    75000   100000
			    150000   200000   275000   350000   450000
			    550000   700000   850000  1000000  1250000
			   1500000  1800000  2100000  2400000  2700000
			   3000000  3500000  4000000  4500000  5000000))

   (legal-effects :initarg :legal-effects
                  :initform '(:quaff :read :eat :create :add-magic :use)
                  :accessor variant.legal-effects)
;;   (object-effects :initarg :object-effects :initform (make-hash-table :test #'equal)
;;		   :accessor variant.object-effects)

   ))


(defclass icfp-board-level (themed-level)
  ((id     :initform "board-level")
   (symbol :initform 'board-level)
   (stores        :initarg :stores     :initform nil  :accessor level.stores)
   (num-stores    :initarg :num-stores :initform 8    :accessor level.num-stores)
   (home-num      :initarg :home-num   :initform 7    :accessor level.home-num))

  (:documentation "The Icfp variant has a special board-level with
stores and special behaviour.  The class is used for dispatching."))

(defclass players-home (house)
  ())

(defclass magic-spell ()
  ((name   :accessor spell.name
	   :initform nil
	   :initarg :name
	   :documentation "Name of the spell.")
   (id     :accessor spell.id
	   :initform nil
	   :initarg :id
	   :documentation "Id for the spell, everyone uses this id.")
   (effect :accessor spell.effect
	   :initform nil :initarg
	   :effect :documentation "A function which is invoked when the spell is cast."))
  (:documentation "A very simple wrapper about the spell and little else."))


(defclass spellbook ()
  ((name   :accessor spellbook.name
	   :initform nil
	   :initarg :name
	   :documentation "The name of the spellbook, used for listings.")
   (id     :accessor spellbook.id
	   :initform nil
	   :initarg :id
	   :documentation "The id for the spellbook, used for lookups, e.g from object-kinds.")
   (size   :accessor spellbook.size
	   :initform 6
	   :initarg :size
	   :documentation "The size of the spellbook, ie how many spells it _can_ take.") 
   (spells :accessor spellbook.spells
	   :initform nil
	   :initarg :spells
	   :documentation "An array of spell-objects."))
  (:documentation "Represents the spell-data of a spellbook, not the actual book."))

(defclass spell-classdata ()
  ((id      :initarg :id
	    :initform nil
	    :accessor spell.id
	    :documentation "Id to the real spell, used for lookup.")
   (level   :initarg :level
	    :initform nil
	    :accessor spell.level
	    :documentation "The level the caster must be.")
   (mana    :initarg :mana
	    :initform nil
	    :accessor spell.mana
	    :documentation "The mana the caster needs.")
   (failure :initarg :failure
	    :initform nil
	    :accessor spell.failure
	    :documentation "The base failure-rate in %.")
   (xp      :initarg :xp
	    :initform nil
	    :accessor spell.xp
	    :documentation "The xp given for first-time casting.")
   (tried   :initarg :tried
	    :initform nil
	    :accessor spell.tried
	    :documentation "This slot is saved with the player-object."))
  
  (:documentation "Information that a class have about a spell.  To cast a spell
the class needs this information.  This information varies from class to class, but
the spell is usually the same."))
   

(defclass spellcasting-class (character-class)
  ((spell-stat      :initarg :spell-stat
		    :initform nil
		    :accessor class.spell-stat
		    :documentation "What is the spell-stat?")
   (spells-at-level :initarg :spells-at-level
		    :initform 1
		    :accessor class.spells-at-level
		    :documentation "At what level does the spellcaster get spells?")
   (spells          :initarg :spells
		    :initform nil
		    :accessor class.spells
		    :documentation "An array with the possible spells (of type spell-classdata) the class can have.")
   (learnt-spells   :initarg :learnt-spells
		    :initform nil
		    :accessor class.learnt-spells
		    :documentation "An array with ids to learnt spells (in order).  Is saved with the
player-object."))
  (:documentation "A subclass of the class character-class.  Represents a class with spell-castin capabilities."))

;; this is an ugly hack, it's used for spell-effects as state-info
(defclass icfp-monster-effect ()
  ((damage     :initform 0        :initarg :damage     :accessor meff.damage)
   (note       :initform nil      :initarg :note       :accessor meff.note)
   (seen       :initform nil      :initarg :seen       :accessor meff.seen)
   (obvious    :initform nil      :initarg :obvious    :accessor meff.obvious)
   (dying-note :initform " dies." :initarg :dying-note :accessor meff.dying-note)
   ))


(defclass black-market (store)
  ()
  (:documentation "A store with steep prices, used as a dispatch class."))

;; change to (col row) format
(defclass icfp-basic-frame-locations (basic-frame-locations)
  ((max-mana :initarg :max-mana :initform '(18 . 0))
   (cur-mana :initarg :cur-mana :initform '(19 . 0))

   (target   :initarg :target  :initform '(20 . 0))
   (cut      :initarg :cut     :initform '(21 . 0))
   (stun     :initarg :stun    :initform '(22 . 0))
   )
  (:documentation "Locations and various settings when printing stuff.
Each location should be a cons with (row . col)."))
  
(defclass icfp-bottom-row-locations (bottom-row-locations)
  ((blind    :initarg :blind    :initform 7)
   (confused :initarg :confused :initform 13)
   (afraid   :initarg :afraid   :initform 22)
   (poisoned :initarg :poisoned :initform 29)
   (study    :initarg :study    :initform 64))
   
  (:documentation "Locations and various settings when printing stuff.
Each location is a fixnum with column in the last row."))


(defgeneric is-spellcaster? (obj)
  (:documentation "Returns T if the object/player is a spellcaster."))

(defgeneric apply-spell-effect! (variant type source target &key x y damage state-object)
  (:documentation "Applies a spell-effect of type TYPE from SOURCE on TARGET at coords (X,Y) for
DAMAGE damage.  The state-object may be used to pass info back to calling function.  The methods
should return NIL or the state-object (with possibly updated state."))

(defgeneric get-melee-weapon (creature)
  (:documentation "Returns the active-object that is used as a melee weapon."))

(defgeneric get-missile-weapon (creature)
  (:documentation "Returns the active-object that is used as missile-weapon."))

(defgeneric get-light-source (creature)
  (:documentation "Returns the light-source used."))

(defgeneric get-charge-status (object)
  (:documentation "Returns the charge-status of the given object or NIL."))

(defgeneric print-cut (variant player settings)
  (:documentation "Prints cut-info according to settings."))

(defgeneric print-stun (variant player settings)
  (:documentation "Prints stun-info according to settings."))

(defgeneric print-poisoned (variant player settings)
  (:documentation "Prints poison-info according to settings."))

(defgeneric print-afraid (variant player settings)
  (:documentation "Prints fear-info according to settings."))

(defgeneric print-confused (variant player settings)
  (:documentation "Prints confusion-info according to settings."))

;;; define relevant object-types for icfp.

(def-obj-type weapon :key <weapon>)
(def-obj-type melee-weapon :is weapon)
(def-obj-type sword :key <sword> :is melee-weapon)
(def-obj-type pole-arm :key <pole-arm> :is melee-weapon)
(def-obj-type hafted :key <hafted> :is melee-weapon)

(def-obj-type missile-weapon :is weapon) ;; clash with bow?
(def-obj-type bow :is missile-weapon :key <bow>
	      :kind-slots ((multiplier :accessor object.multiplier :initform 1 :initarg :multiplier)))
(def-obj-type ammo :key <ammo>)
(def-obj-type digger :is weapon :key <digger>)

(def-obj-type armour :key <armour>)
(def-obj-type body-armour :is armour :key <body-armour>)
(def-obj-type dragonscale-armour :is body-armour :key <dsm-armour>)
(def-obj-type boots :is armour :key <boots>)
(def-obj-type gloves :is armour :key <gloves>)
(def-obj-type shield :is armour :key <shield>)
(def-obj-type headgear :is armour :key <headgear>)
(def-obj-type cloak :is armour :key <cloak>)

(def-obj-type potion :key <potion>)
(def-obj-type money :key <money>)
(def-obj-type scroll :key <scroll>)
(def-obj-type wand :key <wand>)
(def-obj-type staff :key <staff>)
(def-obj-type rod :key <rod>
	      :aobj-slots ((recharge-time :accessor aobj.recharge-time :initform 0 :initarg :recharge-time)))
(def-obj-type book :key <book>)
(def-obj-type spellbook :is book :key <spellbook>)
(def-obj-type prayerbook :is book :key <prayerbook>)
(def-obj-type ring :key <ring>)
(def-obj-type chest :key <chest>)
(def-obj-type light-source :key <light-source>
	      :kind-slots ((status-descs :accessor object.status-descs :initform nil :initarg :status-descs)
			   (max-fuel     :accessor object.max-fuel     :initform nil :initarg :max-fuel)))
	      
(def-obj-type container :key <container>)

(def-obj-type food :key <food>)
(def-obj-type mushroom :is food :key <mushroom>)
(def-obj-type neckwear :key <neckwear>)
(def-obj-type amulet :is neckwear :key <amulet>)

(def-obj-type junk :key <junk>)
(def-obj-type skeleton :is junk :key <skeleton>)


(defun van-make-variant-obj ()
  (make-instance 'icfp-variant
		 :id "langband-icfp"
		 :name "Icfp"

		 :stat-length 6
		 
		 ;; only used by development
		 :sys-file "./variants/icfp/langband-icfp.system"
		 :config-path
		 #+langband-development
		 "./variants/icfp/config"
		 #-langband-development
		 "/var/lib/games/langband/icfp"))


(register-variant& "langband-icfp" #'van-make-variant-obj)
	   
