;; -*- Mode: Lisp; Syntax: Common-Lisp; Package: org.langband.vanilla -*-

#|

Based on:

DESC: variants/vanilla/keys.lisp - assignment of keys
Copyright (c) 2000-2002 - Stig Erik Sand

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

|#

(in-package :org.langband.vanilla)


(setf *current-key-table* *ang-keys*)


(defvar *fallback-bid* 1)
;; 0 means that no bid has been entered so far, but will be
;; interpreted as 1 in the relevant places.  Always non-negative.
(defvar *bid* 0)
;; -1 or 1.
(defvar *bid-sign* 1)

;; Lots of Irritating Superfluous Asterisks.
(defun calculate-real-bid ()
  ;; We might later implement some anti-waste criteria here,
  ;; e.g. always return 1 if there's no robot near us who could cause
  ;; a conflict.
  (* (if (zerop *bid*)
       *fallback-bid*
       *bid*)
     *bid-sign*)
  )

(defun update-bid-display ()
  (assert (not (minusp *bid*)))
  (assert (member *bid-sign* '(-1 1)))
  (when (> (if (zerop *bid*) *fallback-bid* *bid*) *money*)
    (setq *bid* 0 *fallback-bid* 1))
;  (format *debug-io* "~&--- BID=~A~A, real bid=~A~%"
;          (if (plusp *bid-sign*) '+ '-)
;          *bid*
;          (calculate-real-bid))
  (bit-flag-add! *redraw* +print-basic+))

(defun append-to-bid (dungeon player digit)
  (declare (ignore dungeon player))
  (setq *bid* (+ (* *bid* 10) digit))
  (update-bid-display))

(defun execute-action (dungeon player action arg)
  (let ((bid (if (eql action 'Wait)
               ;; It's _never_ useful to spend any money to have the
               ;; first move if we're only going to wait anyway.
               1
               (calculate-real-bid))))
    (send-move-to-server dungeon player bid action arg))
  (unless (zerop *bid*) (setq *fallback-bid* *bid*))
  (setq *bid* 0)
  (update-bid-display)
  (when (> *fallback-bid* (/ *money* 4))
    (format-message! "You are low on money.")))

(define-key-operation 'append-0
    #'(lambda (dungeon player) (append-to-bid dungeon player 0)))
(define-key-operation 'append-1
    #'(lambda (dungeon player) (append-to-bid dungeon player 1)))
(define-key-operation 'append-2
    #'(lambda (dungeon player) (append-to-bid dungeon player 2)))
(define-key-operation 'append-3
    #'(lambda (dungeon player) (append-to-bid dungeon player 3)))
(define-key-operation 'append-4
    #'(lambda (dungeon player) (append-to-bid dungeon player 4)))
(define-key-operation 'append-5
    #'(lambda (dungeon player) (append-to-bid dungeon player 5)))
(define-key-operation 'append-6
    #'(lambda (dungeon player) (append-to-bid dungeon player 6)))
(define-key-operation 'append-7
    #'(lambda (dungeon player) (append-to-bid dungeon player 7)))
(define-key-operation 'append-8
    #'(lambda (dungeon player) (append-to-bid dungeon player 8)))
(define-key-operation 'append-9
    #'(lambda (dungeon player) (append-to-bid dungeon player 9)))
(define-key-operation 'make-bid-positive
    #'(lambda (dungeon player)
        (declare (ignore dungeon player))
        (setq *bid-sign* 1 *fallback-bid* 1)
        (update-bid-display)))
(define-key-operation 'make-bid-negative
    #'(lambda (dungeon player)
        (declare (ignore dungeon player))
        (setq *bid-sign* -1 *fallback-bid* 1)
        (update-bid-display)))
(define-key-operation 'reset-bid
    #'(lambda (dungeon player)
        (declare (ignore dungeon player))
        (setq *bid* 0 *bid-sign* 1 *fallback-bid* 1)
        (update-bid-display)))

;; We _forbid_ suicide by moving into water.

(defun try-moving (dungeon player dir)
  (let ((x (sq-x (robot-sq *self*)))
        (y (sq-y (robot-sq *self*))))
    (ecase (sq-char (get-neighbor (get-sq x y) dir))
      (#\# (format-message! "There is a wall in the way."))
      (#\~ (format-message! "You would drown in the water."))
      ((#\. #\@) (execute-action dungeon player 'Move dir))))
  ;; Update of Langband's data structure will happen when server
  ;; response is parsed.  Also, messages like "You push xxx" will be
  ;; generated then.
  )

(define-key-operation 'move-up
    #'(lambda (dungeon player) (try-moving dungeon player 'N)))

(define-key-operation 'move-right
    #'(lambda (dungeon player) (try-moving dungeon player 'E)))

(define-key-operation 'move-left
    #'(lambda (dungeon player) (try-moving dungeon player 'W)))

(define-key-operation 'move-down
    #'(lambda (dungeon player) (try-moving dungeon player 'S)))

(define-key-operation 'stand-still
    #'(lambda (dungeon player)
        (execute-action dungeon player 'Wait nil)))



(define-key-operation 'quit-game
    #'(lambda (dungeon player)
	(declare (ignore dungeon))
;;	(warn "Quitting")
	(c-prt! "Are you sure you wish to quit? " 0 0)
	(let ((chr (read-one-character)))
	  (when (or (equal chr #\y)
		    (equal chr #\Y))
	    (setf (player.dead-p player) t
		  (player.dead-from player) "quitting"
		  (player.leaving-p player) :quit)
;;	(c-quit! +c-null-value+) ;; how to quit cleanly to the REPL?
	
	    ))
	))


(defun package-input-char (alt-len)
  "INTERNAL FUNCTION.  Might change!

Reads a character via READ-ONE-CHARACTER and
acts on the result:
  Q     - calls QUIT-GAME&
  S     - returns NIL
  ESC   - Picks random value and returns it (a number)
  *     - As ESC
  ENTER - Returns 'CURRENT (ie the currently selected value)
  SPACE - As ENTER
  [a-z] - Checks if the value is legal, returns number if ok, returns 'BAD-VALUE if not legal
"
  (let ((val (read-one-character)))
    #-cmu
    (assert (characterp val))
;;    (warn "Got back ~a ~s ~s" val val (type-of val))
    (cond ((eql val +escape+)
	   'exit)
	  ;; use highlighted value
	  ((or (eql val #\Return) (eql val #\Newline)
	       (eql val #\Space))
	   'current)
	  (t
	   (let ((its-char-code (char-code val)))
	     ;; legal char-code
	     (cond ((and (>= its-char-code (char-code #\a))
			 (<= its-char-code (char-code #\z)))
		    (let ((r-val (- its-char-code (char-code #\a))))
		      (if (and (>= r-val 0) (< r-val alt-len))
			  r-val
			  (progn
			    (c-bell! (format nil "Invalid value: ~a" val))
			    'bad-value))))

		   ;; an arrow-key I guess
		   ((and (>= its-char-code (char-code #\A))
			 (<= its-char-code (char-code #\Z)))
		    (case val
		      (#\N 'up)
		      (#\E 'right)
		      (#\W 'left)
		      (#\S 'down)
		      (t 'bad-value)))

		   (t
		    'bad-value))
	     )))
    ))

(defun package-input-1 (player col row alternatives
		     &key (display-fun nil) (mod-value 5) (bottom-display 10)
		      (base-weight 0)
		       (ask-for "value") (already-selected '()))
  "INTERNAL FUNCTION.
The COL, ROW argument specifies where the alternatives should start
ALTERNATIVES is a list of valid alternatives
DISPLAY-FUN is a function to display help for a given option
MOD-VALUE is how much space should be between rows (I think)
BOTTOM-DISPLAY means something."
  ;; [add more info]

  
  (let ((alt-len (length alternatives))
	(currently-selected already-selected))
    (labels ((display-alternatives (highlight-num)
	       (let ((desc (when display-fun
			     (funcall display-fun highlight-num))))

		 (c-clear-from! bottom-display) ;; clears things
		 
		 (when desc
		   (c-print-text! 2 bottom-display +term-white+ desc :end-col 75)
		   ))
		 
		 (loop for cur-alt in alternatives
		       for i from 0
		       for the-col = (truncate (+ col (* 25 (mod i mod-value))))
		       for the-row = (truncate (+ 1 row (/ i mod-value)))
		       for cur-bg  = (cond ((= i highlight-num)
					    +term-l-blue+)
					   ((find i currently-selected :test #'eql)
					    +term-l-green+)
					   (t
					    +term-white+))
		       do
		       (put-coloured-str! +term-white+ (format nil "~c) " (i2a i)) the-col the-row)
		       (put-coloured-str! cur-bg (format nil " pkg ~d [~d] ~A" (pkg-id cur-alt)
							 (pkg-weight cur-alt) (sq-name-coordinates (pkg-dest cur-alt)))
					  (+ 3 the-col) the-row)
		       ))

	     (toggle-number (num)
	       (cond ((find num currently-selected :test #'eql)
		      (setf currently-selected (remove num currently-selected :test #'eql)))
		     (t
		      (push num currently-selected)))
	       ;;(warn "cursel ~s" currently-selected)
	       (when (> (get-current-weight) *maxweight*)
		 (print-message! "That is too heavy, unselecting.")
		 (setf currently-selected (remove num currently-selected :test #'eql)))
	       )

	     (get-current-weight ()
	       (let ((sum 0))
		 (loop for i in currently-selected
		       do
		       (incf sum (pkg-weight (elt alternatives i))))
		 (+ sum base-weight)))
	     
	     (get-a-value (cur-sel)
	       (display-alternatives cur-sel)
	       (put-coloured-str! +term-l-blue+ (format nil "Weight: ~d of maximum ~d"
							(get-current-weight) *maxweight*)
				  col (1- row))
	       
	       (put-coloured-str! +term-l-red+ (format nil "Choose ~a (~c-~c, ESC to exit): "
						       ask-for (i2a 0) (i2a (- alt-len 1)))
				  col row)
	       
	       (let ((rval (package-input-char alt-len)))
		 (cond ((symbolp rval)
			(case rval
			  (bad-value (get-a-value cur-sel)) ;; retry
			  (current
			   (toggle-number cur-sel)
			   (get-a-value (mod cur-sel alt-len)))
			  
			  (up    (get-a-value (mod (- cur-sel mod-value) alt-len))) ;; move selection
			  (down  (get-a-value (mod (+ cur-sel mod-value) alt-len))) ;; move selection
			  (left  (get-a-value (mod (- cur-sel 1) alt-len))) ;; move selection
			  (right (get-a-value (mod (+ cur-sel 1) alt-len))) ;; move selection
			  (exit
			   ;;(warn "exit")
			   (return-from package-input-1 currently-selected))
			  (t
			   (warn "Unknown symbol returned ~s" rval))))
		       
		       ((integerp rval)
			(toggle-number rval)
			(get-a-value (mod cur-sel alt-len)))
		       
		       ((characterp rval)
			(warn "character ~s back.." rval))
		       (t
			(warn "Odd returns ~s" rval)))

	       )))
      
      (get-a-value 0))))

(defun package-input (player col row alternatives &rest keys
                      &key (ask-for "value") &allow-other-keys)
  (let* ((max-num 26) ; FIXME
         (screens (ceiling (length alternatives) max-num)))
    (loop for screen from 1
          for list = alternatives then (nthcdr max-num alternatives)
          while list
          for num = (min (length list) max-num)
          append (apply #'package-input-1 player col row (subseq list 0 num)
                        :ask-for (format nil "~A (page ~A/~A)"
                                         ask-for screen screens)
                        keys))))



(defun pick-up-packages! (dungeon player)
  (let* ((x (location-x player))
	 (y (location-y player))
	 (square (get-sq x y))
	 (pkgs (sq-pkgs square)))
    
    (unless pkgs
      (print-message! "No packages here!")
      (return-from pick-up-packages! nil))

    ;; we have some pkgs apparently here
    ;;(warn "Possible packages ~s" pkgs)
    (let ((pkgs-passed (if (> (length pkgs) 26)
			  (loop for i from 0 below 26
				for j in pkgs
				collecting j)
			  pkgs))
	  (returned '()))
      (with-new-screen ()
	(c-term-clear!)
	(c-print-text! 2 2 +term-white+
		       #.(concatenate 'string
                               "Please select packages to pick up or drop.  Packages you carry "
                               "have a green colour and package on the floor a white colour. "
                               "Use arrow-keys to move around, and space/enter to toggle selection. "
                               "Hit ESC when you are done."))

	(setf returned (package-input player 1 9 pkgs-passed :mod-value 3
				      :base-weight *weight*
                                      :ask-for "packages to pick up")))
      (let ((which-ids (mapcar #'(lambda (x) (pkg-id (elt pkgs-passed x)))
			       returned)))
	;;(warn "want to pick up packages ~s" which-ids)
	(cond ((eq which-ids nil)
	       (print-message! "No packages picked up!"))
	      (t
	       (execute-action dungeon player 'Pick which-ids)))
        ))

          
    ))

(defun drop-packages! (dungeon player)
  (let* ((carried (robot-pkgs *self*))
	 (returned '())
	 (already-selected (loop for i from 0
				 for j in carried
				 collecting i)))

    (when (null carried)
      (print-message! "You are not carrying any packages!")
      (return-from drop-packages! nil))

    (with-new-screen ()
      (c-term-clear!)
      (c-print-text! 2 2 +term-white+
		     #.(concatenate 'string
				    "Please select packages to pick up or drop.  Packages you carry "
				    "have a green colour and package on the floor a white colour. "
				    "Use arrow-keys to move around, and space/enter to toggle selection. "
				    "Hit ESC when you are done."))
      
      (setf returned (package-input player 1 9 carried :mod-value 3
				    :already-selected already-selected
                                    :ask-for "packages to drop")))
    (let* ((which-ids (mapcar #'(lambda (x) (pkg-id (elt carried x)))
			      returned))
	   (diff (set-difference (mapcar #'pkg-id carried) which-ids)))
	
	;;(warn "want to keep packages ~s, but drop ~s" which-ids diff)
	(cond ((eq diff nil)
	       (print-message! "No packages dropped!"))
	      (t
	       (execute-action dungeon player 'Drop diff)))
        )))


(define-key-operation 'get-item
    #'(lambda (dungeon player)
	(pick-up-packages! dungeon player)))
;;	(with-new-screen ()
;;	(pick-up-from-floor! dungeon player)
	;; ok.. which packages do we have?

(defun show-packages (player)
  (let* ((carried (robot-pkgs *self*)))
    (unless carried
      (print-message! "You don't carry any packages!")
      (return-from show-packages nil))

    
    (with-new-screen ()
      (c-term-clear!)

      (loop for i from 1 below 24
	    for cur-alt in carried
	    for dest = (pkg-dest cur-alt)
	    for dest-x = (square-x dest)
	    for dest-y = (decanonicalize-y (square-y dest))
	    do
	    
	    (put-coloured-str! +term-l-green+
			       (format nil " Package ~4d, destined for (~3d,~3d), weight ~d" (pkg-id cur-alt)
				       dest-x dest-y (pkg-weight cur-alt)) ;; add proper dest later
			       1 (1+ i)))
      (pause-last-line!)
      )))
  

(define-key-operation 'drop-item
    #'(lambda (dungeon player)
	(drop-packages! dungeon player)
;;	(drop-something! dungeon player)
	))

(define-key-operation 'show-help
    #'(lambda (dungeon player)
        (declare (ignore dungeon player))
        (with-new-screen ()
          (c-clear-from! 0)
          (display-help-topics *variant* "Langband help (ICFP)" 3)

          ;;(pause-last-line!)

          )))

(define-key-operation 'show-inventory
    #'(lambda (dungeon player)
	(declare (ignore dungeon))
	(show-packages player)))


(define-key-operation 'redraw-all
    #'(lambda (dungeon player)

	;; skip flushes
	;; do xtra react
	(org.langband.ffi:c-term-xtra& 10 0) ;; xtra_react
	;; skip combine & reorder

	(bit-flag-add! *update* #.(logior +pl-upd-torch+
					  +pl-upd-forget-view+
					  +pl-upd-update-view+ 
					  +pl-upd-monsters+))
	(bit-flag-add! *redraw* #.(logior +print-extra+
					  +print-basic+
					  +print-map+
					  ;; skip equippy
					  ))
	;; do clear
	(c-term-clear!)
	 
	;; call handle-stuff
	(handle-stuff *variant* dungeon player)
	
	;; do all windows, we might need this

	(c-term-fresh!)
	
	))

(define-keypress *ang-keys* :global #\d 'drop-item)
(define-keypress *ang-keys* :global #\g 'get-item)
;;(define-keypress *ang-keys* :global #\i 'show-inventory)
(define-keypress *ang-keys* :global #\Q 'quit-game)

;; Ctrl-R
(define-keypress *ang-keys* :global (code-char 18) 'redraw-all)

(define-keypress *ang-keys* :global #\0 'append-0)
(define-keypress *ang-keys* :global #\1 'append-1)
(define-keypress *ang-keys* :global #\2 'append-2)
(define-keypress *ang-keys* :global #\3 'append-3)
(define-keypress *ang-keys* :global #\4 'append-4)
(define-keypress *ang-keys* :global #\5 'append-5)
(define-keypress *ang-keys* :global #\6 'append-6)
(define-keypress *ang-keys* :global #\7 'append-7)
(define-keypress *ang-keys* :global #\8 'append-8)
(define-keypress *ang-keys* :global #\9 'append-9)
(define-keypress *ang-keys* :global #\- 'make-bid-negative)
(define-keypress *ang-keys* :global #\+ 'make-bid-positive)
(define-keypress *ang-keys* :global #\Backspace 'reset-bid)

(define-keypress *ang-keys* :global #\? 'show-help)
(define-keypress *ang-keys* :global #\i 'show-inventory)

(define-keypress *ang-keys* :global #\. 'stand-still)
(define-keypress *ang-keys* :global #\S 'move-down)
(define-keypress *ang-keys* :global #\W 'move-left)
(define-keypress *ang-keys* :global #\E 'move-right)
(define-keypress *ang-keys* :global #\N 'move-up)


